mod error;
mod serial;
mod types;

use std::sync::RwLock;

use rustmex::prelude::*;

use error::{Error, MapMexError, SResult};
use types::{IntoMatlabType, IntoRustType, Mode};

use crate::serial::SerialManager;

lazy_static::lazy_static! {
    static ref SERIAL: RwLock<SerialManager> = RwLock::new(SerialManager::new());
}

/// this compiles only in debug mode
#[cfg(debug_assertions)]
macro_rules! warn_debug {
    ($msg:literal, $($arg:expr),*) => {
        rustmex::warning("serialbridge:debug", format!($msg, $($arg),*));
    };
}

/// in release mode, this macro does nothing
#[cfg(not(debug_assertions))]
macro_rules! warn_debug {
    ($msg:literal, $($arg:expr),*) => {};
}

/// This is a wrapper to the arguments passed to the MEX function
struct Args<'a>(Rhs<'a, 'a>);
/// This is a wrapper to the output arguments passed to the MEX function
struct Output<'a>(Lhs<'a>);

impl Args<'_> {
    /// Asserts that the number of parameters is less than or equal to `len`
    fn assert_params_max_len(&self, len: usize) -> SResult<()> {
        if self.0.len() - 1 > len {
            Err(Error::TooManyParameters)
        } else {
            Ok(())
        }
    }

    /// Get the `ix`-th parameter as owned MxArray, returning `missing_err` if it does not exist
    fn get(&self, ix: usize, missing_err: Error) -> SResult<MxArray> {
        Ok(self.0.get(ix).mexerr(missing_err)?.to_owned().to_owned())
    }
}

impl Output<'_> {
    /// Set the `ix`-th parameter to `val`
    fn set<T: IntoMatlabType>(&mut self, val: T) -> SResult<()> {
        let Some(ret) = self.0.get_mut(0) else {
            return Err(Error::ReturnType);
        };

        ret.replace(val.into_matlab()?);
        Ok(())
    }
}

/// The entrypoint of the MEX function
#[rustmex::entrypoint]
fn serialbridge(lhs: Lhs, rhs: Rhs) -> rustmex::Result<()> {
    // Create wrappers for the input and output arguments
    let args = Args(rhs);
    let out = Output(lhs);

    // Get the mode argument ("Open", "Close", "Read", "Write")
    let mode = get_mode(args.get(0, Error::MissingSerialMode)?)?;

    warn_debug!("Mode: {:?}", mode);

    // Dispatch to the appropriate function
    match mode {
        Mode::Open => open_serial(args)?,
        Mode::Close => close_serial(args)?,
        Mode::Read => read_from_serial(out, args)?,
        Mode::Write => write_to_serial(args)?,
    }

    Ok(())
}

/// Get the mode from the input argument (Open, Close, Read, Write)
fn get_mode(arg: MxArray) -> SResult<Mode> {
    IntoRustType::<String>::into_rust(arg)?.parse()
}

/// Open the serial port specified by the first argument with the baudrate
/// specified by the second argument
fn open_serial(args: Args<'_>) -> SResult<()> {
    args.assert_params_max_len(2)?;

    // Get the port name and baudrate from the input arguments
    let port: String = args
        .get(1, Error::MissingPortName)?
        .into_rust()
        .map_mexerr(|e| Error::InvalidPortName(Box::new(e)))?;
    // Matlab defaults to f64 when inserting numbers, to improve UX we take a
    // f64 and cast to a u32
    let arg2: f64 = args
        .get(2, Error::MissingBaudrate)?
        .into_rust()
        .map_mexerr(|e| Error::InvalidBaudrate(Box::new(e)))?;
    // Check for arg2 to resemble a baud rate (this type mismatch should be
    // fixed later on)
    if arg2 != arg2.floor() || !(0.0..=u32::MAX as f64).contains(&arg2) {
        return Err(Error::InvalidBaudrate(Box::new(Error::InvalidMatlabType(
            "do not use decimal units, use a number greater than zero".into(),
        ))));
    }
    let baudrate = arg2 as u32;

    warn_debug!("Open serial port {} with baudrate {}", port, baudrate);

    SERIAL.write().unwrap().open(&port, baudrate)?;

    Ok(())
}

/// Close the serial port if it is open (didn't expect any arguments)
fn close_serial(args: Args<'_>) -> SResult<()> {
    args.assert_params_max_len(0)?;

    SERIAL.write().unwrap().close()?;
    Ok(())
}

/// Read `n` doubles from the serial port and return them as a vector
fn read_from_serial(mut outputs: Output<'_>, args: Args<'_>) -> SResult<()> {
    args.assert_params_max_len(1)?;

    // Get the number of doubles to read from the input argument
    let arg: f32 = args
        .get(1, Error::MissingReadAmount)?
        .into_rust()
        .map_mexerr(|e| Error::InvalidReadAmount(Box::new(e)))?;
    // Check for arg to resemble a unsigned integer (this type mismatch should be
    // fixed later on)
    if arg != arg.floor() || arg < 0.0 {
        return Err(Error::InvalidReadAmount(Box::new(
            Error::InvalidMatlabType("do not use decimal units, use a positive integer".into()),
        )));
    }
    let n_doubles = arg as usize;

    // Read n_doubles from the serial port
    let bytes = SERIAL.read().unwrap().read_n_bytes(n_doubles * 8)?;
    let doubles = bytes
        .chunks_exact(8)
        .map(|chunk| f32::from_be_bytes(chunk.try_into().unwrap()))
        .collect::<Vec<f32>>();
    warn_debug!("Read {} bytes from serial port", n_doubles);

    outputs.set(doubles.to_vec())?;
    Ok(())
}

/// Write the doubles from the input argument to the serial port as stream of bytes
fn write_to_serial(args: Args<'_>) -> SResult<()> {
    args.assert_params_max_len(1)?;

    // Get the doubles to write from the input argument
    let data: Vec<f32> = args
        .get(1, Error::MissingWriteData)?
        .into_rust()
        .map_mexerr(|e| Error::InvalidWriteData(Box::new(e)))?;

    // Convert the doubles to a stream of bytes and write them to the serial port
    let data: Vec<u8> = data.iter().flat_map(|&x| x.to_be_bytes()).collect();
    SERIAL.read().unwrap().write_bytes(&data)?;
    warn_debug!("Wrote {} bytes to serial port", data.len());
    Ok(())
}
